#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#define MAX_CONNECTIONS 32
#define BUFFER 4096

void eerror(char* msg);
int server(int port);
void run(int serverfd);
void textrun(int serverfd);
int main(int argc, char* argv[]) {
	int port = 8888;
	int tmp = 0;
	if (argc > 1) {
		tmp = atoi(argv[argc-1]);
		if (tmp) {
			port = tmp;
		}
	}
	if (argc > 2 || argv[0][0] == 't' || argv[0][0]=='T') {
		textrun(server(port));
	} else {
		run(server(port));
	}
	return 0;
}
int server(int port) {
	int connfd,serverfd;
	struct sockaddr_in serveraddr;
	serverfd = socket(AF_INET,SOCK_STREAM,0);
	bzero(&serveraddr,sizeof(serveraddr));
	serveraddr.sin_family = AF_INET;
	serveraddr.sin_port = htons(port);
	serveraddr.sin_addr.s_addr = htonl(INADDR_ANY);
	bind(serverfd, (struct sockaddr *)&serveraddr,sizeof(serveraddr));
	listen(serverfd,5);
	fprintf(stderr,"Waiting for connection on port %d\n",port);
	return serverfd;
}
void eerror(char* msg) {
	perror(msg);
	exit(1);
}
/*int  select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout);*/
/*
 *Bugs Disconnects half way through etc.
 *
 *
 *
 */
int findMax(int * read_fd,int serverfd,int count) {
	int i = 0;
	int max = -1;
	for (i = 0; i < count; i++ ) {
		if (read_fd[i] > max) {
			max = read_fd[i];
		}
	}	
	if (max < serverfd) {
		max = serverfd;
	}
	return max;
}
int addFd(int * read_fd,int count,int value) {
	int i = 0;
	for (i = 0; i < count; i++ ) {
		if (read_fd[i] <= 0) {
			read_fd[i] = value;
			return i;
		}
	}	
	return -1;
}
int read_from_client (int filedes,short * in,short * out,int divisor,int size)
{
	int nbytes;
	int sizeBytes = size*sizeof(short);
	int i;

	nbytes = read (filedes, in, sizeBytes);
	if (nbytes < 0) {
		/* Read error. */
		perror ("read");
		exit (EXIT_FAILURE);
	} else if (nbytes == 0) {
		return -1;
	} else {
		/* Data read. */
		if (nbytes < sizeBytes) {
			bzero(in+nbytes/sizeof(short),sizeBytes-nbytes);
		}
		for (i = 0; i < size; i++) {
			out[i] +=  in[i]/divisor;
		}
		/* fprintf (stderr, "Server: got message: `%s'\n", buffer);*/
		return 0;
	}
	return 0;
}

void run(int serverfd) {
	int * read_fd =(int *) malloc(sizeof(int) * MAX_CONNECTIONS);
	short * inBuffer =(short *) malloc(sizeof(short) * BUFFER);
	short * outBuffer =(short*) malloc(sizeof(short) * BUFFER);
	fd_set active_fd_set;
	fd_set read_fd_set;
	struct sockaddr_in clientname;
	int max = serverfd;
	int count = 0;
	int currCount = 0;
	int i;
	FD_ZERO (&active_fd_set);
	FD_ZERO (&read_fd_set);
	FD_SET (serverfd, &active_fd_set);
	

	bzero(read_fd,sizeof(int)*MAX_CONNECTIONS);
	bzero(outBuffer,sizeof(short)*BUFFER);
	for (;;) {
		/* Block until input arrives on one or more active sockets. */
		read_fd_set = active_fd_set; /* COPY */
		currCount = select (max+1, &read_fd_set, NULL, NULL, NULL);
		if (currCount < 0) {
			perror ("select");
			exit (EXIT_FAILURE);
		}
		if (FD_ISSET(serverfd,&read_fd_set)) {
			/* Connection request on original socket. */
			int new;
			int size = sizeof (struct sockaddr);
			new = accept(serverfd, (struct sockaddr *) &clientname, &size);
			if (new < 0) {
				perror ("accept");
				exit (EXIT_FAILURE);
			}
			fprintf (stderr, "Server: connect from host %s, port %hd.\n", inet_ntoa (clientname.sin_addr), ntohs (clientname.sin_port));
			FD_SET (new, &active_fd_set);
			if (addFd(read_fd,MAX_CONNECTIONS,new) < 0) {
				perror ("Too Many Connections");
				exit (EXIT_FAILURE);
			}
			count++;
			max = findMax(read_fd,serverfd,MAX_CONNECTIONS);
			currCount--;

		}

		/* Service all the sockets with input pending. */
		bzero(outBuffer,sizeof(short)*BUFFER);
		for (i = 0; i < MAX_CONNECTIONS; i++) {
			if (read_fd[i] > 0 && FD_ISSET (read_fd[i], &read_fd_set)) {
				/* Data arriving on an already-connected socket. */
				fprintf(stderr,".");
				if (read_from_client(read_fd[i],inBuffer,outBuffer,count,BUFFER) < 0) {
					close (read_fd[i]);
					FD_CLR (read_fd[i], &active_fd_set);
					max = findMax(read_fd,serverfd,MAX_CONNECTIONS);
					count--;
				}
			}
		} /* for */
		write(STDOUT_FILENO,outBuffer,(sizeof(short)*BUFFER));	
		/* now outbuffer is done */
	} /* for (;;) */
}
/* why did I copy and paste? LAZYNESS! */
void textrun(int serverfd) {
	int * read_fd =(int *) malloc(sizeof(int) * MAX_CONNECTIONS);
	FILE ** read_file =(FILE**) malloc(sizeof(FILE*) * MAX_CONNECTIONS);
	short * inBuffer =(short *) malloc(sizeof(short) * BUFFER);
	short * outBuffer =(short*) malloc(sizeof(short) * BUFFER);
	fd_set active_fd_set;
	fd_set read_fd_set;
	struct sockaddr_in clientname;
	int max = serverfd;
	int count = 0;
	int currCount = 0;
	int i;
	FD_ZERO (&active_fd_set);
	FD_ZERO (&read_fd_set);
	FD_SET (serverfd, &active_fd_set);
	

	bzero(read_fd,sizeof(int)*MAX_CONNECTIONS);
	bzero(outBuffer,sizeof(short)*BUFFER);
	for (;;) {
		/* Block until input arrives on one or more active sockets. */
		read_fd_set = active_fd_set; /* COPY */
		currCount = select (max+1, &read_fd_set, NULL, NULL, NULL);
		if (currCount < 0) {
			perror ("select");
			exit (EXIT_FAILURE);
		}
		if (FD_ISSET(serverfd,&read_fd_set)) {
			/* Connection request on original socket. */
			int new;
			int size = sizeof (struct sockaddr);
			new = accept(serverfd, (struct sockaddr *) &clientname, &size);
			if (new < 0) {
				perror ("accept");
				exit (EXIT_FAILURE);
			}
			fprintf (stderr, "Server: connect from host %s, port %hd.\n", inet_ntoa (clientname.sin_addr), ntohs (clientname.sin_port));
			FD_SET (new, &active_fd_set);
			i = addFd(read_fd,MAX_CONNECTIONS,new);
			if (i < 0) {
				perror ("Too Many Connections");
				exit (EXIT_FAILURE);
			}
			read_file[i] = fdopen(read_fd[i],"w");
			count++;
			max = findMax(read_fd,serverfd,MAX_CONNECTIONS);
			currCount--;

		}

		/* Service all the sockets with input pending. */
		bzero(outBuffer,sizeof(short)*BUFFER);
		for (i = 0; i < MAX_CONNECTIONS; i++) {
			if (read_fd[i] > 0 && FD_ISSET (read_fd[i], &read_fd_set)) {
				/* Data arriving on an already-connected socket. */
				fprintf(stderr,".");
				if (fgets(inBuffer,BUFFER,read_file[i])==NULL) {
					fclose(read_file[i]);
					read_file[i] = NULL;
					close (read_fd[i]);
					FD_CLR (read_fd[i], &active_fd_set);
					max = findMax(read_fd,serverfd,MAX_CONNECTIONS);
					count--;
				} else {
					fputs(inBuffer,stdout);
				}
			}
		} /* for */
	} /* for (;;) */
}


